﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace SunCRC.Crc.Crc1_64
{
    class Crc1_64Model
    {
        public ulong Polynomial { get; private set; }
        public ulong Initial { get; private set; }
        public ulong FinalXor { get; private set; }
        public bool InputReflected { get; private set; }
        public bool ResultReflected { get; private set; }

        public int CrcSize { get; private set; }

        public ulong TopBitMask { get; private set; }

        public int ShiftByteLeftBits { get; private set; }

        public ulong FinalMask { get; private set; }

        public int BitOffsetTo8Bits { get; private set; }

        public string Name { get; private set; }

        public ulong OriginalPolynomial { get; private set; }

        public Crc1_64Model() { }
        public Crc1_64Model(int crcSize, ulong polynomial, ulong initialVal, bool inputReflected, bool resultReflected,
                            ulong finalXorVal, string name)
        {
            this.Polynomial = polynomial;
            this.OriginalPolynomial = polynomial;       /* save polynomial for later if refelcted model is requested */
            this.Initial = initialVal;
            this.FinalXor = finalXorVal;
            this.InputReflected = inputReflected;
            this.ResultReflected = resultReflected;
            this.Name = name;

            if (crcSize > 64)
            {
                throw new ArgumentException("Only Crc with size in range [1, 64] are supported");
            }

            this.CrcSize = crcSize;
            this.TopBitMask = (ulong)(1UL << (CrcSize - 1));
            this.FinalMask = GetFinalMask(CrcSize);

            // if the crcsize is less than 8, we still work with an 8-bit shift register and use the MSB as top bit mask because as input
            // always a byte is processed which makes it easier
            // in general, the XOR operation is applied only to with the "x" (where x = [1, 7]) upper bits of the shift register for CRC size < 8.
            if (crcSize < 8)
            {
                int diffgap = 8 - crcSize;

                this.TopBitMask = 0x80;
                this.FinalMask = 0xFF;
                this.ShiftByteLeftBits = 0;

                this.Polynomial = this.Polynomial << diffgap;
                this.BitOffsetTo8Bits = diffgap;
            }
            else
            {
                this.ShiftByteLeftBits = crcSize - 8;
                this.BitOffsetTo8Bits = 0;
            }
        }

        public static Crc1_64Model GetReflectedCrcModel(Crc1_64Model model)
        {
            ulong initial = CrcUtil.ReflectGeneral(model.Initial, model.CrcSize);
            bool inputReflected = !model.InputReflected;
            bool resultReflected = !model.ResultReflected;

            // use original polynomial because actual polynomial might have been changed in constructor
            ulong polynomial = CrcUtil.ReflectGeneral(model.OriginalPolynomial, model.CrcSize);

            ulong finalmask = GetFinalMask(model.CrcSize);

            Crc1_64Model reflectedModel = new Crc1_64Model();
            reflectedModel.CrcSize = model.CrcSize;
            reflectedModel.Polynomial = polynomial;
            reflectedModel.Initial = initial;
            reflectedModel.InputReflected = inputReflected;
            reflectedModel.ResultReflected = resultReflected;
            reflectedModel.FinalXor = model.FinalXor;
            reflectedModel.Name = model.Name;
            reflectedModel.FinalMask = finalmask;

            reflectedModel.TopBitMask = 0; // not used
            reflectedModel.ShiftByteLeftBits = 0; // not used
            reflectedModel.BitOffsetTo8Bits = 0;

            return reflectedModel;
        }

        private static ulong GetFinalMask(int crcSize)
        {
            ulong finalMask;
            // workaround for CRC64 because 1 << 64 overflow, so set mask directly
            if (crcSize == 64)
            {
                finalMask = 0xFFFFFFFFFFFFFFFFUL;
            }
            else
            {
                finalMask = (ulong)(1UL << crcSize) - 1;
            }
            return finalMask;
        }
    }
}
